@@ -1,12 +1,13 @@
 """
 Functionality to set, serve from, and clear the cache.
 """
-
+import re
 from enum import Enum
 from functools import wraps
-from typing import Callable, Optional
-
+from typing import Callable, Optional, List
+from urllib.parse import unquote
 from django.core.cache import caches
+from django.core.cache.backends.base import BaseCache
 from django.core.handlers.wsgi import WSGIRequest
 from django.http.response import HttpResponse
 from django.template.response import SimpleTemplateResponse
@@ -18,7 +19,7 @@ from django.utils.cache import (
     patch_response_headers,
 )
 from django.utils.deprecation import MiddlewareMixin
-from wagtail.core import hooks
+from wagtail import hooks
 
 from wagtailcache.settings import wagtailcache_settings
 
@@ -51,6 +52,50 @@ def _patch_header(response: HttpResponse, status: Status) -> None:
         response[wagtailcache_settings.WAGTAIL_CACHE_HEADER] = status.value
 
 
+def _chop_querystring(r: WSGIRequest) -> WSGIRequest:
+    """
+    Given a request object, remove any of our ignored querystrings from it.
+    """
+    if len(r.GET) and wagtailcache_settings.WAGTAIL_CACHE_IGNORE_QS:
+        # Make a copy of querystrings, and delete any that should be ignored.
+        qs = r.GET.copy()
+        for q in r.GET:
+            for regex in wagtailcache_settings.WAGTAIL_CACHE_IGNORE_QS:
+                if re.match(regex, q):
+                    del qs[q]
+        # Mutate the request to include our chopped up querystrings. We must
+        # also mutate the raw QUERY_STRING as that is used within
+        # ``request.build_absolute_uri()`` which is used in Django cache
+        # middleware internals.
+        r.GET = qs
+        r.META["QUERY_STRING"] = qs.urlencode()
+    return r
+
+
+def _get_cache_key(r: WSGIRequest, c: BaseCache) -> str:
+    """
+    Wrapper for Django's get_cache_key which first strips specific
+    querystrings. Since the Django cache middleware is somewhat complicated and
+    RFC compliant, we are best off to chop and pass it along rather than
+    re-inventing the cache keying logic.
+    """
+    r = _chop_querystring(r)
+    return get_cache_key(r, None, r.method, c)
+
+
+def _learn_cache_key(
+    r: WSGIRequest, s: HttpResponse, t: int, c: BaseCache
+) -> str:
+    """
+    Wrapper for Django's learn_cache_key which first strips specific
+    querystrings. Since the Django cache middleware is somewhat complicated and
+    RFC compliant, we are best off to chop and pass it along rather than
+    re-inventing the cache keying logic.
+    """
+    r = _chop_querystring(r)
+    return learn_cache_key(r, s, t, None, c)
+
+
 class FetchFromCacheMiddleware(MiddlewareMixin):
     """
     Loads a request from the cache if it exists.
@@ -88,21 +133,14 @@ class FetchFromCacheMiddleware(MiddlewareMixin):
             setattr(request, "_wagtailcache_skip", True)
             return None  # Don't bother checking the cache.
 
-        # Try and get the cached GET response.
-        cache_key = get_cache_key(request, None, "GET", cache=self._wagcache)
+        # Try and get the cached response.
+        cache_key = _get_cache_key(request, self._wagcache)
         if cache_key is None:
             setattr(request, "_wagtailcache_update", True)
             return None  # No cache information available, need to rebuild.
 
         response = self._wagcache.get(cache_key)
 
-        # If it wasn't found and a HEAD was requested, try looking for that.
-        if response is None and request.method == "HEAD":
-            cache_key = get_cache_key(
-                request, None, "HEAD", cache=self._wagcache
-            )
-            response = self._wagcache.get(cache_key)
-
         if response is None:
             setattr(request, "_wagtailcache_update", True)
             return None  # No cache information available, need to rebuild.
@@ -180,9 +218,23 @@ class UpdateCacheMiddleware(MiddlewareMixin):
             timeout = self._wagcache.default_timeout
         patch_response_headers(response, timeout)
         if timeout:
-            cache_key = learn_cache_key(
-                request, response, timeout, None, cache=self._wagcache
+            cache_key = _learn_cache_key(
+                request, response, timeout, self._wagcache
             )
+
+            # Track cache keys based on URI.
+            # (of the chopped request, not the real one).
+            cr = _chop_querystring(request)
+            uri = unquote(cr.build_absolute_uri())
+            keyring = self._wagcache.get("keyring", {})
+            # Get current cache keys belonging to this URI.
+            # This should be a list of keys.
+            uri_keys: List[str] = keyring.get(uri, [])
+            # Append the key to this list and save.
+            uri_keys.append(cache_key)
+            keyring[uri] = uri_keys
+            self._wagcache.set("keyring", keyring)
+
             if isinstance(response, SimpleTemplateResponse):
                 response.add_post_render_callback(
                     lambda r: self._wagcache.set(cache_key, r, timeout)
@@ -196,12 +248,43 @@ class UpdateCacheMiddleware(MiddlewareMixin):
         return response
 
 
-def clear_cache() -> None:
+def clear_cache(urls: List[str] = None) -> None:
     """
-    Clears the entire cache backend used by wagtail-cache.
+    Clears the Wagtail cache backend.
+
+    :param urls: An optional list of strings, representing regular expressions
+    to match against the list of URLs in the cache. If a URL matches, it is
+    deleted from the cache. If ``urls`` is ``None`` the entire cache is cleared.
     """
-    if wagtailcache_settings.WAGTAIL_CACHE:
-        caches[wagtailcache_settings.WAGTAIL_CACHE_BACKEND].clear()
+
+    if not wagtailcache_settings.WAGTAIL_CACHE:
+        return
+
+    _wagcache = caches[wagtailcache_settings.WAGTAIL_CACHE_BACKEND]
+    if urls and "keyring" in _wagcache:
+        keyring = _wagcache.get("keyring")
+
+        # Check the provided URL matches a key in our keyring.
+        matched_urls = []
+        for regex in urls:
+            for key in keyring:
+                if re.match(regex, key):
+                    matched_urls.append(key)
+
+        # If it matches, delete each entry from the cache,
+        # and delete the URL from the keyring.
+        for url in matched_urls:
+            entries = keyring.get(url, [])
+            for cache_key in entries:
+                _wagcache.delete(cache_key)
+            del keyring[url]
+
+        # Save the keyring.
+        _wagcache.set("keyring", keyring)
+
+    # Clears the entire cache backend used by wagtail-cache.
+    else:
+        _wagcache.clear()
 
 
 def cache_page(view_func: Callable[..., HttpResponse]):
